﻿using System;
using System.Diagnostics;
using System.ServiceProcess;
using System.Net.Sockets;
using System.Net;
using System.IO;
using System.Text;
using System.ComponentModel;

namespace Ips_Get_Tcp_Stream
{
    public partial class Ips_Get_Tcp_Stream : ServiceBase
    {
        public Ips_Get_Tcp_Stream()
        {
            InitializeComponent();
        }

        //Method that runs when the service receives a "start" command.
        protected override void OnStart(string[] args)
        {
            //A thread is created and started, which will run the method "listenToStream()" .
            System.Threading.Thread t = new System.Threading.Thread(listenToStream);
            t.Start();
        }        

        //Method that runs when the service receives a "stop" command.
        protected override void OnStop()
        {
            ServiceController service = new ServiceController("Ips_Get_Tcp_Stream");
            try
            {
                //Tries to stop the service
                TimeSpan timeout = TimeSpan.FromMilliseconds(10000);
                service.Stop();
                service.WaitForStatus(ServiceControllerStatus.Stopped, timeout);
            }
            catch
            {
                File.AppendAllText(@"F:\Database\Database_HMN_Production\TrackingData\LogFiles\errorlog.txt", "Error - " + DateTime.Now.ToString("yyyyMMddHHmm") + "Service failed to stop." + Environment.NewLine);
            }
        }

        public static void listenToStream()
        {
            //A buffer is set for the data comming through the TCP-stream.
            int BufferSize = 10000;
            //A TcpListener that listens to the IP of the server this service is running on. The port has been set to 55200, which is the port the data is being sent to, from the Cisco MSE.
            TcpListener listener = new TcpListener(IPAddress.Parse(Program.getIP()), 55200);
            //"ReuseAddress" ensures that the possibility to listen to a port even if an other application already is using it.
            listener.Server.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
            //Variable for the Client sending data.
            TcpClient client;
            //Service starts listening for incomming connections to the IP and port specified for the TcpListener.
            listener.Start();

            //This loop runs continuously.
            //It accept all incomming connections and saves the data to F:\Database\Database_HMN_Production\TrackingData\rawData
            //Then it checks if the files contains 1 or 2 xml-files, and if so, saves them in F:\Database\Database_HMN_Production\TrackingData\processedData\   
            //It also removes any text above or under the xml files.         
            while (true)
            {
                //Accept any client trying to send data.
                client = listener.AcceptTcpClient();
                //Grabs the clients stream.
                NetworkStream tcpStream = client.GetStream();
                //Sets a byte-buffer for the incomming data.
                byte[] buffer = new byte[BufferSize];
                //String variable to save the converted data.
                string message = "";
                //Count the number of bytes received.
                int bytesReceived;

                //Get and convert the incomming data from bytes to string.
                do
                {
                    bytesReceived = tcpStream.Read(buffer, 0, client.ReceiveBufferSize);
                    message += Encoding.ASCII.GetString(buffer, 0, bytesReceived);
                }
                while (bytesReceived > 0);

                //Try to process the data
                try {
                    //Sets paths for files.
                    string time = DateTime.Now.ToString("yyyyMMddHHmmssffffff");
                    string filename = @"F:\Database\Database_HMN_Production\TrackingData\rawData\TcpStreamLog" + time;
                    string filename_processed = @"F:\Database\Database_HMN_Production\TrackingData\processedData\TcpStreamLog" + time;

                    //To avoid multiple xmls being saved in the same file, a duplicate check is performed here, and "_2" is added if file already exists.
                    if (File.Exists(filename + ".xml"))
                    {
                        filename = filename + "_2.xml";
                        File.AppendAllText(filename, message);
                    }
                    else
                    {
                        filename = filename + ".xml";
                        File.AppendAllText(filename, message);
                    }

                    //Temporary files are used to save data before moving the files to \processedData                        
                    string tempFile = @"F:\Database\Database_HMN_Production\TrackingData\rawData\Temp1";                    
                    string tempFile2 = @"F:\Database\Database_HMN_Production\TrackingData\rawData\Temp2";

                    //Clear temp files before use                        
                    System.IO.File.WriteAllText(tempFile, string.Empty);                    
                    System.IO.File.WriteAllText(tempFile2, string.Empty);

                    //Booleans are used to know if we have one or more complete xml-files in the incomming data.
                    bool foundStart = false;
                    bool foundEnd = false;
                    bool foundSecondXml = false;
                    bool foundSecondEnd = false;

                    //Here we loop through the incomming data, which has been saved as "filename".
                    //If there is an xml-file in the data, this will be saved to a temporary file "tempFile".
                    //If there is a second xml-file, this will be saved to another temporary file "tempFile2".
                    using (var sr = new StreamReader(filename))
                    using (var sw = new StreamWriter(tempFile))
                    using (var sw2 = new StreamWriter(tempFile2))
                    {
                        string line;
                        while ((line = sr.ReadLine()) != null)
                        {
                            if (foundStart == true)
                            {
                                if (foundEnd == false)
                                {
                                    sw.WriteLine(line);
                                    if (line.Contains("</StreamingNotification>"))
                                    {
                                        foundEnd = true;
                                    }
                                }
                                //add eventual extra xml to other file.
                                else if (foundEnd == true)
                                {
                                    if (line.Contains("<?xml"))
                                    {
                                        foundSecondXml = true;
                                        sw2.WriteLine(line);
                                    }
                                    else if (foundSecondXml == true)
                                    {
                                        if (line.Contains("</StreamingNotification>"))
                                        {
                                            foundSecondEnd = true;
                                            sw2.WriteLine(line);
                                        }
                                        else if (foundSecondEnd == true)
                                        {
                                            sr.ReadLine();
                                        }
                                        else
                                        {
                                            sw2.WriteLine(line);
                                        }
                                    }
                                }
                            }
                            else if (line.Contains("<?xml"))
                            {
                                foundStart = true;
                                sw.WriteLine(line);
                            }
                            else
                            {
                                sr.ReadLine();
                            }
                        }
                    }

                    //Check if 1 or more xmls has been found in the file, in order to save them if so.
                    //Sometimes files arrive rapidly, and to avoid conflicting use of the same filename, "_2" is added to the filename if it already exists.
                    if (foundEnd == true)
                    {
                        if (File.Exists(filename_processed + ".xml"))
                        {
                            File.Move(tempFile, filename_processed + "_2.xml");
                            if (foundSecondEnd == true)
                            {
                                File.Move(tempFile2, filename_processed + "_2dbl.xml");
                            }
                        }
                        else
                        {
                            File.Move(tempFile, filename_processed + ".xml");
                            if (foundSecondEnd == true)
                            {
                                File.Move(tempFile2, filename_processed + "dbl.xml");
                            }
                        }
                        //************************************************************************************************
                        //Import to the database could be implemented here, after the files has been processed and saved.
                        //************************************************************************************************
                        
                    }
                    else
                    {
                        File.AppendAllText(@"F:\Database\Database_HMN_Production\TrackingData\LogFiles\errorlog.txt", DateTime.Now.ToString("yyyyMMddHHmm") + "Could not find end of XML  -   filename: " + filename + Environment.NewLine);
                    }
                }
                catch
                {
                    File.AppendAllText(@"F:\Database\Database_HMN_Production\TrackingData\LogFiles\errorlog.txt", DateTime.Now.ToString("yyyyMMddHHmm") + "Failed to process data" + Environment.NewLine);
                }
            }
        }
    }
}